# pylint: disable=no-member
"""
Fueltech updates

Revision ID: 76b7e7c9462a
Revises: 2ed53b064d85
Create Date: 2021-02-25 05:54:05.861570

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "76b7e7c9462a"
down_revision = "2ed53b064d85"
branch_labels = None
depends_on = None

FUELTECHS = {
    "AGLNOW1": "bioenergy_biomass",
    "AGLSITA1": "bioenergy_biomass",
    "APPIN": "gas_recip",
    "AWABAREF": "bioenergy_biomass",
    "BBASEHOS": "distillate",
    "BERWICK": "bioenergy_biomass",
    "BPLANDF1": "bioenergy_biomass",
    "BROADMDW": "bioenergy_biomass",
    "BROOKLYN": "bioenergy_biomass",
    "BWTR1": "bioenergy_biomass",
    "CLAYTON": "bioenergy_biomass",
    "CONDONG1": "bioenergy_biomass",
    "CORIO1": "bioenergy_biomass",
    "EASTCRK": "bioenergy_biomass",
    "EASTCRK2": "bioenergy_biomass",
    "EDLRGNRD": "bioenergy_biomass",
    "GERMCRK": "gas_recip",
    "GLENNCRK": "gas_recip",
    "GRANGEAV": "bioenergy_biomass",
    "GROSV1": "gas_recip",
    "GROSV2": "gas_recip",
    "HALAMRD1": "bioenergy_biomass",
    "HIGHBRY1": "bioenergy_biomass",
    "JACKSGUL": "bioenergy_biomass",
    "KINCUM1": "bioenergy_biomass",
    "LUCAS2S2": "bioenergy_biomass",
    "LUCASHGT": "bioenergy_biomass",
    "MBAHNTH": "gas_recip",
    "MORNW": "bioenergy_biomass",
    "OAKY2": "gas_recip",
    "PEDLER1": "bioenergy_biomass",
    "REMOUNT": "bioenergy_biomass",
    "ROCHEDAL": "bioenergy_biomass",
    "SHEP1": "bioenergy_biomass",
    "STAPYLTON1": "bioenergy_biomass",
    "SVALE1": "bioenergy_biomass",
    "TAHMOOR1": "gas_recip",
    "TATURA01": "bioenergy_biomass",
    "TEATREE1": "bioenergy_biomass",
    "TERALBA": "gas_recip",
    "TITREE": "bioenergy_biomass",
    "TOWER": "gas_recip",
    "WDLNGN01": "bioenergy_biomass",
    "WHIT1": "bioenergy_biomass",
    "WINGF1_1": "bioenergy_biomass",
    "WINGF2_1": "bioenergy_biomass",
    "WOLLERT1": "bioenergy_biomass",
    "WOYWOY1": "bioenergy_biomass",
    "WYNDW": "bioenergy_biomass",
}


def get_update_sql(code: str, fueltech_id: str) -> str:
    return f"update facility set fueltech_id='{fueltech_id}' where code='{code}'"


def upgrade() -> None:
    for facility_code, fueltech_id in FUELTECHS.items():
        op.execute(get_update_sql(facility_code, fueltech_id))


def downgrade() -> None:
    pass



"""
OpenNEM Facility Capacities


"""

from textwrap import dedent
from typing import Dict

from opennem.db import get_database_engine
from opennem.schema.network import NetworkSchema


def facility_capacity_network_region_fueltech_query(
    network: NetworkSchema, network_region: str
) -> str:
    query = """
    select
        f.fueltech_id,
        sum(f.capacity_registered)
    from facility f
    where
        f.network_id='{network_id}'
        and f.network_region='{network_region}'
        and f.active is True
    group by 1
    order by 1;
    """.format(
        network_id=network.code, network_region=network_region
    )

    return dedent(query)


def get_facility_capacities(network: NetworkSchema, network_region: str) -> Dict:
    engine = get_database_engine()

    query = facility_capacity_network_region_fueltech_query(
        network=network, network_region=network_region
    )
    results = []

    with engine.connect() as c:
        results = list(c.execute(query))

    return_dict = {i[0]: i[1] for i in results}

    return return_dict



# pylint: disable=no-self-argument
"""
OpenNEM Energy Tools

Uses an average in 30 minute buckets

@TODO variable bucket sizes

"""
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkSchema

logger = logging.getLogger("opennem.compat.energy")


class ScadaResultCompat(BaseConfig):
    interval: datetime
    code: Optional[str]
    generated: Union[float, int, None]


def _trapezium_integration_variable(d_ti: pd.Series) -> float:
    """Gapfill version of trap int - will fill out"""
    # Clear no numbers
    d_ti = d_ti.dropna()

    # Fall back on average but warn to check data
    if d_ti.count() <= 3:
        d_sum = d_ti.sum()

        if d_sum == 0:
            return 0

        if d_ti.count() == 0:
            return 0

        return 0.5 * d_sum / d_ti.count()

    bucket_middle = d_ti.count() - 2

    bucket_middle_weights = [1] + [2] * bucket_middle + [1]

    weights = d_ti.values * bucket_middle_weights

    weights_sum = weights.sum()

    bucket_energy = 0.5 * weights_sum / ((d_ti.count() - 1) * 2)

    return bucket_energy


def _trapezium_integration(d_ti: pd.Series, gapfill: bool = False) -> Optional[float]:
    """ Energy for a 30 minute bucket """
    # Clear no numbers
    d_ti = d_ti.dropna()

    # Early return
    if d_ti.sum() == 0:
        return 0.0

    # Fall back on average but warn to check data
    if d_ti.count() != 7:
        # logger.error("Series {} has gaps".format(d_ti))

        if gapfill:
            return _trapezium_integration_variable(d_ti)

        return None

    weights = d_ti.values * [1, 2, 2, 2, 2, 2, 1]

    weights_sum: float = weights.sum()

    bucket_energy = 0.5 * weights_sum / 12

    return bucket_energy


def _energy_aggregate(df: pd.DataFrame, network: NetworkSchema) -> pd.DataFrame:
    """Energy aggregate that buckets into 30min intervals
    but takes edges in for 7 total values"""
    in_cap = {}
    capture: Dict[str, Any] = {}
    values = []
    dt: Optional[datetime] = None
    reading_stops: List[int] = [0, 30]

    for index, value in df.iterrows():
        i, network_id, duid = index

        if duid not in capture:
            capture[duid] = []
            in_cap[duid] = True

        if in_cap[duid]:
            capture[duid].append(value.eoi_quantity)

            if dt:
                energy = _trapezium_integration(pd.Series(capture[duid]), True)
                values.append((dt, network_id, duid, energy))

            in_cap[duid] = False
            capture[duid] = []

        if i.minute in reading_stops:
            dt = i
            capture[duid].append(value.eoi_quantity)
            in_cap[duid] = True

        else:
            capture[duid].append(value.eoi_quantity)

    return pd.DataFrame(
        values, columns=["trading_interval", "network_id", "facility_code", "eoi_quantity"]
    )


def energy_sum(gen_series: List[Dict], network: NetworkSchema) -> pd.DataFrame:
    """Takes the energy sum for a series of raw duid intervals
    and returns a fresh dataframe to be imported"""
    df = pd.DataFrame(
        gen_series,
        columns=[
            "trading_interval",
            "facility_code",
            "network_id",
            "eoi_quantity",
        ],
    )

    # Clean up types
    df.trading_interval = pd.to_datetime(df.trading_interval)
    df.eoi_quantity = pd.to_numeric(df.eoi_quantity)

    # Index by datetime
    df = df.set_index(["trading_interval", "network_id", "facility_code"])

    # Multigroup by datetime and facility code
    df = _energy_aggregate(df, network=network)

    # Reset back to a simple frame
    df = df.reset_index()

    # Shift back 5 minutes to even up into 30 minute bkocks
    if network.reading_shift:
        df.trading_interval = df.trading_interval - pd.Timedelta(minutes=network.reading_shift)

    # Add back the timezone for NEM
    # We use a fixed offset so need to loop
    df.trading_interval = df.apply(
        lambda x: pd.Timestamp(x.trading_interval, tz=network.get_fixed_offset()), axis=1
    )

    # filter out empties
    df = df[pd.isnull(df.eoi_quantity) == False]

    return df



"""
Multiprocessor for energy workers

"""
import logging
import multiprocessing
import sys
from datetime import datetime
from typing import List, Optional, Tuple

import click

from opennem.core.facility.fueltechs import load_fueltechs
from opennem.diff.versions import CUR_YEAR, get_network_regions
from opennem.schema.network import NetworkNEM
from opennem.workers.energy import run_energy_update_archive

logger = logging.getLogger("opennem.worker.energy_multi")

CPU_COUNT = multiprocessing.cpu_count()
WORKER_COUNT = 2
YEAR_EARLIEST = 2020
CUR_MONTH = datetime.now().month


def _build_args_list(
    year: Optional[int], region: Optional[str], fueltech: Optional[str]
) -> List[Tuple[Optional[int], int, Optional[str], Optional[str]]]:
    args_list = []

    network = NetworkNEM

    regions = [region]

    if not region:
        regions = [i.code for i in get_network_regions(network)]

    fueltechs = [fueltech]

    if not fueltech:
        fueltechs = [i for i in load_fueltechs().keys()]

    years = [year]

    if not year:
        years = [i for i in range(CUR_YEAR, YEAR_EARLIEST - 1, -1)]

    for year in years:
        for month in range(12, 0, -1):
            if year == CUR_YEAR and month > CUR_MONTH:
                continue

            for region in regions:
                if year == CUR_YEAR and month > CUR_MONTH:
                    continue

                for ft in fueltechs:
                    args_list.append((year, month, region, ft))

    return args_list


def _worker_wrap(year: Optional[int], months: int, region: str, fueltech_id: Optional[str]) -> None:
    """Map to named args"""
    return run_energy_update_archive(
        year=year, months=[months], regions=[region], fueltech_id=fueltech_id
    )


def energy_process(args_list: List, worker_count: Optional[int]) -> None:
    logger.debug("Starting with {} workers".format(WORKER_COUNT))

    if not worker_count:
        worker_count = WORKER_COUNT

    with multiprocessing.Pool(processes=worker_count) as pool:
        pool.starmap(_worker_wrap, args_list)


@click.command()
@click.option("--year", required=False, type=int)
@click.option("--workers", required=False, type=int)
@click.option("--fueltech", required=False, type=str)
@click.option("--region", required=False, type=str)
def cli(
    year: Optional[int] = None,
    fueltech: Optional[str] = None,
    region: Optional[str] = None,
    workers: Optional[int] = None,
) -> None:
    args_list = _build_args_list(year=year, region=region, fueltech=fueltech)
    click.echo("Running {} items".format(len(args_list)))
    energy_process(args_list, worker_count=workers)


if __name__ == "__main__":
    try:
        cli()
    except KeyboardInterrupt:
        logger.error("User stopped")
        sys.exit(-1)
    except Exception as e:
        logger.error(e)
